# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import shutil

import os
import platform
import pytest

from unittest import mock

from callee import Contains
from .conftest import git_out, search_diff, search_rev

from mozphab import environment, exceptions, mozphab

call_conduit = mock.Mock()


def by_line_mock(*args, **_kwargs):
    # join args to catch unicode errors
    " ".join(*args)
    return ["Revision URI: http://example.test/D123"]


check_call_by_line = mock.Mock()
check_call_by_line.side_effect = by_line_mock

initial_sha = None


def test_submit_create(in_process, git_repo_path, init_sha):
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # user search
        [dict(userName="alice", phid="PHID-USER-1")],
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    (git_repo_path / "X").write_text("ą\nb\nc\n", encoding="utf-8")
    (git_repo_path / "Y").write_text("no line ending")
    git_out("add", ".")
    (git_repo_path / "msg").write_text("Ą r?alice", encoding="utf-8")
    git_out("commit", "--file", "msg")
    (git_repo_path / "untracked").write_text("a\n")

    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - Ą r?alice

Differential Revision: http://example.test/D123

"""
    assert log.strip() == expected.strip()
    assert mock.call("conduit.ping", {}) in call_conduit.call_args_list
    assert (
        mock.call("user.query", dict(usernames=["alice"]))
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "diffusion.repository.search",
            dict(limit=1, constraints=dict(callsigns=["TEST"])),
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.creatediff",
            {
                "sourceControlPath": "/",
                "sourceControlSystem": "git",
                "lintStatus": "none",
                "sourcePath": mock.ANY,
                "unitStatus": "none",
                "sourceMachine": "http://example.test",
                "sourceControlBaseRevision": mock.ANY,
                "repositoryPHID": "PHID-REPO-1",
                "branch": "HEAD",
                "changes": [
                    {
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "newProperties": {"unix:filemode": "100644"},
                        "oldPath": None,
                        "hunks": [
                            {
                                "oldOffset": 0,
                                "oldLength": 0,
                                "newOffset": 1,
                                "newLength": 3,
                                "addLines": 3,
                                "delLines": 0,
                                "corpus": "+ą\n+b\n+c\n",
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                            }
                        ],
                        "oldProperties": {},
                        "currentPath": "X",
                        "fileType": 1,
                        "type": 1,
                        "metadata": {},
                    },
                    {
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "newProperties": {"unix:filemode": "100644"},
                        "oldPath": None,
                        "hunks": [
                            {
                                "oldOffset": 0,
                                "oldLength": 0,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 1,
                                "delLines": 0,
                                "corpus": (
                                    "+no line ending\n\\ No newline at end of file\n"
                                ),
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": True,
                            }
                        ],
                        "oldProperties": {},
                        "currentPath": "Y",
                        "fileType": 1,
                        "type": 1,
                        "metadata": {},
                    },
                ],
                "creationMethod": "moz-phab-git",
            },
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.setdiffproperty",
            {"diff_id": "1", "name": "local:commits", "data": ~Contains('"rev":')},
        )
        in call_conduit.call_args_list
    )
    assert (
        call_conduit.call_args_list.count(
            mock.call("differential.setdiffproperty", mock.ANY)
        )
        == 2
    )


def test_submit_create_added_not_commited(in_process, git_repo_path, init_sha):
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # user search
        [dict(userName="alice", phid="PHID-USER-1")],
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    (git_repo_path / "X").write_text("ą\r\nb\nc\n", encoding="utf-8")
    (git_repo_path / "Y").write_text("no line ending")
    git_out("add", ".")
    (git_repo_path / "msg").write_text("Ą r?alice", encoding="utf-8")
    git_out("commit", "--file", "msg")
    (git_repo_path / "untracked").write_text("a\n")
    git_out("add", "untracked")

    with pytest.raises(exceptions.Error) as excinfo:
        mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    assert "Uncommitted changes present." in str(excinfo.value)


def test_submit_create_no_bug(in_process, git_repo_path, init_sha):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # user search
        [dict(userName="alice", phid="PHID-USER-1")],
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    testfile = git_repo_path / "X"
    testfile.write_text("a\n")
    git_out("add", ".")
    msgfile = git_repo_path / "msg"
    msgfile.write_text("A r?alice")
    git_out("commit", "--file", "msg")

    mozphab.main(["submit", "--yes", "--no-bug", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
A r?alice

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()


def test_submit_create_binary(in_process, git_repo_path, init_sha, data_file):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # file.allocate
        dict(dict(filePHID=None, upload=True)),
        # file.upload
        dict(),
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    shutil.copyfile(str(data_file), str(git_repo_path / "img.png"))
    git_out("add", ".")
    git_out("commit", "-m", "IMG")

    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - IMG

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "file.allocate",
            {"name": "img.png", "contentHash": mock.ANY, "contentLength": 182},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call("file.upload", {"data_base64": mock.ANY, "name": "img.png"})
        in call_conduit.call_args_list
    )


def test_submit_create_binary_existing(in_process, git_repo_path, init_sha, data_file):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # file.allocate
        dict(dict(filePHID="PHID-FILE-1", upload=False)),
        # no file.upload call
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    shutil.copyfile(str(data_file), str(git_repo_path / "img.png"))
    git_out("add", ".")
    git_out("commit", "-m", "IMG")

    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - IMG

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "file.allocate",
            {"name": "img.png", "contentHash": mock.ANY, "contentLength": 182},
        )
        in call_conduit.call_args_list
    )
    assert mock.call("file.upload", mock.ANY) not in call_conduit.call_args_list


def test_submit_create_binary_chunked(in_process, git_repo_path, init_sha, data_file):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # file.allocate
        dict(dict(filePHID="PHID-FILE-1", upload=True)),
        # file.querychunks
        [
            dict(byteStart="0", byteEnd="4194304", complete=False),
            dict(byteStart="4194304", byteEnd="8388608", complete=False),
            dict(byteStart="8388608", byteEnd="8425160", complete=False),
        ],
        # file.uploadchunk
        dict(),
        # file.uploadchunk
        dict(),
        # file.uploadchunk
        dict(),
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    shutil.copyfile(str(data_file), str(git_repo_path / "img.png"))
    git_out("add", ".")
    git_out("commit", "-m", "IMG")

    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - IMG

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "file.allocate",
            {"name": "img.png", "contentHash": mock.ANY, "contentLength": 182},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call("file.querychunks", {"filePHID": "PHID-FILE-1"})
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "file.uploadchunk",
            {
                "filePHID": "PHID-FILE-1",
                "byteStart": 0,
                "data": mock.ANY,
                "dataEncoding": "base64",
            },
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "file.uploadchunk",
            {
                "filePHID": "PHID-FILE-1",
                "byteStart": 4194304,
                "data": mock.ANY,
                "dataEncoding": "base64",
            },
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "file.uploadchunk",
            {
                "filePHID": "PHID-FILE-1",
                "byteStart": 8388608,
                "data": mock.ANY,
                "dataEncoding": "base64",
            },
        )
        in call_conduit.call_args_list
    )


def test_submit_update(in_process, git_repo_path, init_sha):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # diffusion.revision.search
        dict(data=[search_rev(rev=123)]),
        # diffusion.diff.search
        dict(data=[search_diff()]),
        # whoami
        dict(phid="PHID-USER-1"),
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    testfile = git_repo_path / "X"
    testfile.write_text("ą", encoding="utf-8")
    git_out("add", ".")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1 - Ą

Differential Revision: http://example.test/D123
""",
        encoding="utf-8",
    )
    git_out("commit", "--file", "msg")

    mozphab.main(
        ["submit", "--yes"]
        + ["--bug", "1"]
        + ["--message", "update message ćwikła"]
        + [init_sha],
        is_development=True,
    )

    assert call_conduit.call_count == 9
    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """\
Bug 1 - Ą

Differential Revision: http://example.test/D123

"""
    assert log == expected


@mock.patch("mozphab.commands.submit.logger.warning")
def test_submit_update_no_change(
    m_logger_warning, in_process, git_repo_path, init_sha, git_sha
):
    testfile = git_repo_path / "X"
    testfile.write_text("a")
    git_out("add", ".")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1 - A, r=test

Differential Revision: http://example.test/D123
"""
    )
    git_out("commit", "--file", "msg")
    sha = git_sha()
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # diffusion.revision.search
        dict(data=[search_rev(rev=123, reviewers=("test",))]),
        # diffusion.diff.search
        dict(data=[search_diff(node=sha)]),
        # whoami
        dict(phid="PHID-USER-1"),
        # user.query
        [dict(phid="PHID-USER-2", userName="test")],
    )

    mozphab.main(
        ["submit", "--yes"] + [init_sha],
        is_development=True,
    )
    assert call_conduit.call_count == 6
    assert m_logger_warning.called_once_with("No changes to submit.")


def test_submit_remove_cr(in_process, git_repo_path, init_sha):
    if environment.IS_WINDOWS:
        pytest.skip("Removing CR will not work on Windows.")

    call_conduit.side_effect = (
        # CREATE
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # user.search
        [dict(userName="alice", phid="PHID-USER-1")],
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
        # UPDATE
        # no need to ping (checked)
        # no need to check reviewer
        # no need to search for repository repository data is saved in .hg
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-2", diffid="2")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="124")),
        # differential.setdiffproperty
        dict(),
    )
    test_a = git_repo_path / "X"
    test_a.write_text("a\r\nb\n")
    git_out("add", "X")
    git_out("commit", "-am", "A r?alice")
    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)
    call_conduit.reset_mock()
    # removing CR, leaving LF
    test_a.write_text("a\nb\n")
    git_out("commit", "-am", "B r?alice")
    mozphab.main(["submit", "--yes", "--bug", "1", "HEAD~"], is_development=True)

    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 2,
                        "fileType": 1,
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 2,
                                "newOffset": 1,
                                "newLength": 2,
                                "addLines": 1,
                                "delLines": 1,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": "-a\r\n+a\n b\n",
                            }
                        ],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "git",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-git",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "HEAD",
            },
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.setdiffproperty",
            {
                "diff_id": "2",
                "name": "local:commits",
                "data": Contains('"summary": "Bug 1 - B r?alice"')
                & Contains(
                    '"message": "'
                    "Bug 1 - B r?alice\\n\\n"
                    "Summary:\\n\\n\\n\\n\\n"
                    "Test Plan:\\n\\n"
                    "Reviewers: alice\\n\\n"
                    "Subscribers:\\n\\n"
                    'Bug #: 1"'
                ),
            },
        )
        in call_conduit.call_args_list
    )


def test_submit_remove_form_feed(in_process, git_repo_path, init_sha):
    """Test deleting a file with a form feed character will not corrupt the diff."""
    call_conduit.side_effect = (
        # CREATE
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # user.search
        [dict(userName="alice", phid="PHID-USER-1")],
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )

    test_a = git_repo_path / "X"
    test_a.write_text("some line\fcontaining form feed\na second line\n")
    git_out("add", "X")
    git_out("commit", "-am", "A r?alice")

    # delete file
    test_a.unlink()
    git_out("commit", "-am", "B r?alice")

    call_conduit.reset_mock()
    mozphab.main(["submit", "--yes", "--bug", "1", "HEAD~"], is_development=True)

    failing = True
    for arg in call_conduit.call_args_list:
        if arg[0][0] == "differential.creatediff":
            assert arg[0][1]["changes"][0]["hunks"][0]["corpus"] == (
                "-some line\fcontaining form feed\n-a second line\n"
            )
            failing = False

    assert not failing


def test_submit_single_last(in_process, git_repo_path, init_sha):
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    (git_repo_path / "X").write_text("a\n")
    git_out("add", "X")
    git_out("commit", "-am", "A")
    (git_repo_path / "X").write_text("b\n")
    git_out("commit", "-am", "B")

    mozphab.main(["submit", "--yes", "--bug", "1", "--single"], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-2")
    expected = """\
Bug 1 - B

Differential Revision: http://example.test/D123
A


"""
    assert log == expected


def test_submit_single_first(in_process, git_repo_path, init_sha, git_sha):
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    (git_repo_path / "X").write_text("a\n")
    git_out("add", "X")
    git_out("commit", "-am", "A")
    sha = git_sha()
    (git_repo_path / "X").write_text("b\n")
    git_out("commit", "-am", "B")

    mozphab.main(
        ["submit", "--yes", "--bug", "1", "--single", sha], is_development=True
    )

    log = git_out("log", "--format=%s%n%n%b", "-2")
    expected = """\
B


Bug 1 - A

Differential Revision: http://example.test/D123
"""
    assert log == expected


def test_submit_update_no_message(in_process, git_repo_path, init_sha):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        dict(data=[search_rev(rev=123)]),
        dict(data=[search_diff()]),
        dict(phid="PHID-USER-1"),
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    (git_repo_path / "X").write_text("ą", encoding="utf-8")
    git_out("add", ".")
    (git_repo_path / "msg").write_text(
        """\
Bug 1 - Ą

Differential Revision: http://example.test/D123
""",
        encoding="utf-8",
    )
    git_out("commit", "--file", "msg")

    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """\
Bug 1 - Ą

Differential Revision: http://example.test/D123

"""
    assert log == expected


def test_submit_update_revision_not_found(in_process, git_repo_path, init_sha):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # response for searching D123 and D124
        dict(data=[search_rev(rev=123)]),
        dict(data=[search_diff()]),
        # moz-phab asks again for D124
        dict(data=[]),
        # whoami
        dict(phid="PHID-USER-1"),
        # moz-phab asks again for D124
        dict(data=[]),
        # moz-phab asks again for D124
        dict(data=[]),
    )
    testfile = git_repo_path / "X"
    testfile.write_text("ą", encoding="utf-8")
    git_out("add", ".")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1 - Ą

Differential Revision: http://example.test/D123
""",
        encoding="utf-8",
    )
    git_out("commit", "--file", "msg")
    testfile.write_text("missing repo")
    msgfile.write_text(
        """\
Bug 1 - missing revision

Differential Revision: http://example.test/D124
"""
    )
    git_out("commit", "--all", "--file", "./msg")

    with pytest.raises(exceptions.Error) as excinfo:
        mozphab.main(
            ["submit", "--yes"]
            + ["--bug", "1"]
            + ["--message", "update message ćwikła"]
            + [init_sha],
            is_development=True,
        )
    assert "query result for revision D124" in str(excinfo.value)


def test_empty_file(in_process, git_repo_path, init_sha):
    # Add an empty file
    call_conduit.side_effect = (
        # ping
        dict(),
        # diffusion.repository.search
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-1", diffid="1")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="123")),
        # differential.setdiffproperty
        dict(),
    )
    testfile = git_repo_path / "X"
    testfile.touch()
    git_out("add", ".")
    git_out("commit", "--message", "A")

    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - A

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": None,
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {"unix:filemode": "100644"},
                        "commitHash": mock.ANY,
                        "type": 1,
                        "fileType": 1,
                        "hunks": [],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "git",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-git",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "HEAD",
            },
        )
        in call_conduit.call_args_list
    ), (
        "The diff should have populated fields and "
        "`change.hunks` should be an empty list"
    )

    # Modify a file's mode
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-2", diffid="2")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="124")),
        # differential.setdiffproperty
        dict(),
    )
    # Windows cannot update UMASK bits in the same way Unix
    # systems can, so use Git's update-index command to
    # directly modify the index for the file.
    if platform.system() == "Windows":
        git_out("update-index", "--chmod=+x", str(testfile))
    else:
        os.chmod(testfile, 0o0755)
    git_out("commit", "-a", "--message", "B")
    mozphab.main(["submit", "--yes", "--bug", "1", "HEAD~"], is_development=True)
    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - B

Differential Revision: http://example.test/D124
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {"unix:filemode": "100644"},
                        "newProperties": {"unix:filemode": "100755"},
                        "commitHash": mock.ANY,
                        "type": 2,
                        "fileType": 1,
                        "hunks": [],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "git",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-git",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "HEAD",
            },
        )
        in call_conduit.call_args_list
    ), (
        "The diff should contain filemode changes and "
        "`change.hunks` should be an empty list"
    )

    # Remove an empty file
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # differential.creatediff
        dict(dict(phid="PHID-DIFF-3", diffid="3")),
        # differential.setdiffproperty
        dict(),
        # differential.revision.edit
        dict(object=dict(id="125")),
        # differential.setdiffproperty
        dict(),
    )
    testfile.unlink()
    git_out("commit", "-a", "--message", "C")
    mozphab.main(["submit", "--yes", "--bug", "1", "HEAD~"], is_development=True)
    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - C

Differential Revision: http://example.test/D125
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {"unix:filemode": "100755"},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 3,
                        "fileType": 1,
                        "hunks": [],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "git",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-git",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "HEAD",
            },
        )
        in call_conduit.call_args_list
    ), (
        "The diff should have empty `newProperties` and "
        "`change.hunks` should be an empty list"
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from unittest import mock
import pytest

from .conftest import git_out, hg_out

from mozphab import exceptions, mozphab


call_conduit = mock.Mock()


def test_no_need_to_reorganise(in_process, git_repo_path, init_sha):
    # One commit
    call_conduit.side_effect = (
        dict(),  # ping
        dict(data=[dict(phid="PHID-1", id=1)]),  # differential.get_revision
        # Revision is not related to any other revision. There is no stack.
        dict(data=[]),  # edge.search
    )

    f = git_repo_path / "X"
    f.write_text("A")
    git_out("add", ".")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1: A r?alice

Differential Revision: http://example.test/D1
"""
    )
    git_out("commit", "--file", "msg")
    with pytest.raises(exceptions.Error) as e:
        mozphab.main(["reorg", "--yes", init_sha], is_development=True)

    assert (str(e.value)) == "Reorganisation is not needed."

    # Stack of commits
    call_conduit.side_effect = (
        dict(data=[dict(phid="PHID-1", id=1), dict(phid="PHID-2", id=2)]),
        # PHID-2 is the only child of PHID-1.
        dict(
            data=[
                dict(
                    sourcePHID="PHID-1",
                    destinationPHID="PHID-2",
                    edgeType="revision.child",
                )
            ]
        ),
    )

    f.write_text("B")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1: B r?alice

Differential Revision: http://example.test/D2
"""
    )
    git_out("commit", "-a", "--file", "msg")

    with pytest.raises(exceptions.Error) as e:
        mozphab.main(["reorg", "--yes", init_sha], is_development=True)

    assert (str(e.value)) == "Reorganisation is not needed."


def test_new_separate_revisions_to_stack(in_process, git_repo_path, init_sha):
    call_conduit.side_effect = (
        # ping
        dict(),
        # search revisions
        dict(
            data=[
                dict(
                    phid="PHID-1",
                    id=1,
                    fields=dict(status=dict(value="needs-review")),
                ),
                dict(
                    phid="PHID-2",
                    id=2,
                    fields=dict(status=dict(value="needs-review")),
                ),
            ]
        ),
        # edge search
        dict(data=[]),
        # differential.edit_revision
        dict(data=[dict(phid="PHID-1", id=1)]),
    )

    f = git_repo_path / "X"
    f.write_text("A")
    git_out("add", ".")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1: A r?alice

Differential Revision: http://example.test/D1
"""
    )
    git_out("commit", "--file", "msg")
    f.write_text("B")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1: B r?alice

Differential Revision: http://example.test/D2
"""
    )
    git_out("commit", "-a", "--file", "msg")
    mozphab.main(["reorg", "--yes", init_sha], is_development=True)
    assert (
        mock.call(
            "differential.revision.edit",
            {
                "objectIdentifier": "PHID-1",
                "transactions": [{"type": "children.set", "value": ["PHID-2"]}],
            },
        )
        in call_conduit.call_args_list
    )


def test_add_revision_existing_stack(in_process, git_repo_path, init_sha):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # search revisions
        dict(
            data=[
                dict(
                    phid="PHID-1",
                    id=1,
                    fields=dict(status=dict(value="needs-review")),
                ),
                dict(
                    phid="PHID-2",
                    id=2,
                    fields=dict(status=dict(value="needs-review")),
                ),
                dict(
                    phid="PHID-3",
                    id=3,
                    fields=dict(status=dict(value="needs-review")),
                ),
            ]
        ),
        # edge search
        dict(
            data=[
                dict(
                    sourcePHID="PHID-1",
                    destinationPHID="PHID-2",
                    edgeType="revision.child",
                )
            ]
        ),  # differential.edge.search
        dict(
            data=[
                dict(
                    sourcePHID="PHID-1",
                    destinationPHID="PHID-2",
                    edgeType="revision.parent",
                )
            ]
        ),  # differential.edge.search
        # differential.edit_revision
        dict(data=[dict(phid="PHID-1", id=1)]),
        dict(data=[dict(phid="PHID-3", id=1)]),
        dict(data=[dict(phid="PHID-3", id=1)]),
    )

    f = git_repo_path / "X"
    f.write_text("A")
    git_out("add", ".")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1: A r?alice

Differential Revision: http://example.test/D1
"""
    )
    git_out("commit", "--file", "msg")
    fn = git_repo_path / "Y"
    fn.write_text("C")
    git_out("add", ".")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1: C r?alice

Differential Revision: http://example.test/D3
"""
    )
    git_out("commit", "--file", "msg")
    f.write_text("B")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1: B r?alice

Differential Revision: http://example.test/D2
"""
    )
    git_out("commit", "-a", "--file", "msg")
    mozphab.main(["reorg", "--yes", init_sha], is_development=True)
    assert (
        mock.call(
            "differential.revision.edit",
            {
                "objectIdentifier": "PHID-1",
                "transactions": [{"type": "children.set", "value": ["PHID-3"]}],
            },
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.revision.edit",
            {
                "objectIdentifier": "PHID-3",
                "transactions": [{"type": "children.set", "value": ["PHID-2"]}],
            },
        )
        in call_conduit.call_args_list
    )


def test_add_revision_existing_stack_hg(in_process, hg_repo_path):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        dict(),
        # search revisions
        dict(
            data=[
                dict(
                    phid="PHID-1",
                    id=1,
                    fields=dict(status=dict(value="needs-review")),
                ),
                dict(
                    phid="PHID-2",
                    id=2,
                    fields=dict(status=dict(value="needs-review")),
                ),
                dict(
                    phid="PHID-3",
                    id=3,
                    fields=dict(status=dict(value="needs-review")),
                ),
            ]
        ),
        # edge search
        dict(
            data=[
                dict(
                    sourcePHID="PHID-1",
                    destinationPHID="PHID-2",
                    edgeType="revision.child",
                ),
                dict(
                    sourcePHID="PHID-1",
                    destinationPHID="PHID-3",
                    edgeType="revision.child",
                ),
            ]
        ),  # differential.edge.search
        dict(
            data=[
                dict(
                    sourcePHID="PHID-1",
                    destinationPHID="PHID-2",
                    edgeType="revision.parent",
                )
            ]
        ),  # differential.edge.search
        dict(
            data=[
                dict(
                    sourcePHID="PHID-1",
                    destinationPHID="PHID-3",
                    edgeType="revision.parent",
                )
            ]
        ),  # differential.edge.search
    )

    f = hg_repo_path / "X"
    f.write_text("A")
    hg_out("add")
    msgfile = hg_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1: A r?alice

Differential Revision: http://example.test/D1
"""
    )
    hg_out("commit", "-l", "msg")
    fn = hg_repo_path / "Y"
    fn.write_text("C")
    hg_out("add")
    msgfile = hg_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1: C r?alice

Differential Revision: http://example.test/D3
"""
    )
    hg_out("commit", "-l", "msg")
    f.write_text("B")
    msgfile = hg_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1: B r?alice

Differential Revision: http://example.test/D2
"""
    )
    hg_out("commit", "-l", "msg")
    with pytest.raises(exceptions.Error) as e:
        mozphab.main(["reorg", "--yes", "1", "3"], is_development=True)

    assert (str(e.value)) == "Revision D1 has multiple children."


def test_abandon_a_revision(in_process, git_repo_path, init_sha):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        dict(),  # ping
        dict(
            data=[
                dict(
                    phid="PHID-1",
                    id=1,
                    fields=dict(status=dict(value="needs-review")),
                )
            ]
        ),  # differential.revision.search
        dict(
            data=[
                dict(
                    sourcePHID="PHID-1",
                    destinationPHID="PHID-2",
                    edgeType="revision.child",
                )
            ]
        ),  # differential.edge.search
        dict(
            data=[
                dict(
                    sourcePHID="PHID-1",
                    destinationPHID="PHID-2",
                    edgeType="revision.parent",
                )
            ]
        ),  # differential.edge.search
        dict(
            data=[
                dict(
                    phid="PHID-2",
                    id=2,
                    fields=dict(status=dict(value="needs-review")),
                )
            ]
        ),  # differential.revision.search
        dict(data=[dict(phid="PHID-1", id=1)]),  # differential.edit_revision
        dict(data=[dict(phid="PHID-2", id=2)]),  # differential.edit_revision
    )

    f = git_repo_path / "X"
    f.write_text("A")
    git_out("add", ".")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1: A r?alice

Differential Revision: http://example.test/D1
"""
    )
    git_out("commit", "--file", "msg")
    mozphab.main(["reorg", "--yes", init_sha], is_development=True)
    # Search for the revision to get its PHID
    assert call_conduit.call_args_list[1] == mock.call(
        "differential.revision.search",
        {"constraints": {"ids": [1]}, "attachments": {"reviewers": True}},
    )
    # Search for direct related revisions of PHID-2
    assert call_conduit.call_args_list[2] == mock.call(
        "edge.search",
        {
            "sourcePHIDs": ["PHID-1"],
            "types": ["revision.parent", "revision.child"],
            "limit": 10000,
        },
    )
    # Search for direct related revisions of PHID-1
    assert call_conduit.call_args_list[3] == mock.call(
        "edge.search",
        {
            "sourcePHIDs": ["PHID-2"],
            "types": ["revision.parent", "revision.child"],
            "limit": 10000,
        },
    )
    # Search for revisions in the stack
    assert call_conduit.call_args_list[4] == mock.call(
        "differential.revision.search",
        {"constraints": {"phids": ["PHID-2"]}, "attachments": {"reviewers": True}},
    )
    # Remove the child from PHID-1 and abandon PHID-1
    assert call_conduit.call_args_list[5] == mock.call(
        "differential.revision.edit",
        {
            "transactions": [{"type": "children.remove", "value": ["PHID-2"]}],
            "objectIdentifier": "PHID-1",
        },
    )
    assert call_conduit.call_args_list[6] == mock.call(
        "differential.revision.edit",
        {
            "transactions": [{"type": "abandon", "value": True}],
            "objectIdentifier": "PHID-2",
        },
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import mimetypes
import os
import re
import subprocess
import sys
import uuid

from typing import (
    List,
    Optional,
)

from datetime import datetime
from functools import lru_cache

from mozphab import environment

from .config import config
from .diff import Diff
from .exceptions import CommandError, Error, NotFoundError
from .gitcommand import GitCommand
from .helpers import (
    prompt,
    short_node,
    temporary_binary_file,
    temporary_file,
    create_hunk_lines,
    is_valid_email,
)
from .logger import logger
from .repository import Repository
from .spinner import wait_message
from .telemetry import telemetry

NULL_SHA1 = "0" * 40


class Git(Repository):
    def __init__(self, path):
        dot_path = os.path.join(path, ".git")
        if not os.path.exists(dot_path):
            raise ValueError("%s: not a git repository" % path)

        logger.debug("found git repo in %s", path)

        self.git = GitCommand()

        if os.path.isfile(dot_path):
            # We're working from a worktree. Let's find the dot_path directory.
            dot_path = self.git_out(
                ["rev-parse", "--git-common-dir"], path=path, split=False
            )

        super().__init__(path, dot_path)

        self.vcs = "git"
        m = re.search(r"\d+\.\d+\.\d+", self.git.output(["--version"], split=False))
        if not m:
            raise Error("Failed to determine Git version.")

        self.vcs_version = m.group(0)
        self.revset = None
        self.branch = None

    @property
    def is_cinnabar_installed(self):
        """Check if Cinnabar extension is callable."""
        return self.git.is_cinnabar_installed

    @property
    def is_cinnabar_required(self):
        """Check if local VCS is different than the remote one."""
        return self.vcs != self.phab_vcs

    def _hg_to_git(self, node):
        """Convert Mercurial hashtag to Git."""
        if not self.is_cinnabar_required:
            return None

        return self.git_out(["cinnabar", "hg2git", node], split=False)

    @lru_cache(maxsize=None)
    def _git_to_hg(self, node):
        """Convert Git hashtag to Mercurial."""
        if not self.is_cinnabar_required:
            return None

        hg_node = self.git_out(["cinnabar", "git2hg", node], split=False)
        return hg_node if hg_node != NULL_SHA1 else None

    @lru_cache(maxsize=128)
    def get_public_node(self, node):
        """Return a Mercurial node if Cinnabar is required."""
        public_node = node
        if self.is_cinnabar_required:
            hg_node = self._git_to_hg(node)
            if hg_node:
                public_node = hg_node

        return public_node

    def is_index_modified(self):
        """Are there any changes added to the staging area."""
        return bool(self.git.output(["diff-index", "HEAD"]))

    def is_worktree_clean(self):
        return all(
            [l.startswith("?? ") for l in self.git_out(["status", "--porcelain"])]
        )

    def before_submit(self):
        self.validate_email()

        if self.is_index_modified():
            raise Error(
                "Uncommitted changes present. "
                "Please stash them or commit before submitting."
            )

        # Store current branch (fails if HEAD in detached state)
        try:
            self.branch = self._get_current_head()
        except Exception:
            raise Error(
                "Git failed to read the branch name.\n"
                "The repository is in a detached HEAD state.\n"
                "You need to run the *git checkout <branch-name>* command."
            )

    @classmethod
    def is_repo(cls, path):
        """Quick check for repository at specified path."""
        return os.path.exists(os.path.join(path, ".git"))

    def git_call(self, command, **kwargs):
        """Call git from the repository path."""
        self.git.call(command, cwd=self.path, **kwargs)

    def git_out(self, command, path=None, extra_env=None, **kwargs):
        """Call git from the repository path and return the result."""
        return self.git.output(
            command, cwd=path or self.path, extra_env=extra_env, **kwargs
        )

    def cleanup(self):
        self.git_call(["gc", "--auto", "--quiet"])
        if self.branch:
            self.checkout(self.branch)

    def _find_branches_to_rebase(self, commits):
        """Create a list of branches to rebase."""
        branches_to_rebase = dict()
        for commit in commits:
            if commit["node"] == commit["orig-node"]:
                continue
            branches = self.git_out(["branch", "--contains", commit["orig-node"]])
            for branch in branches:
                if branch.startswith("* ("):
                    # Omit `* (detached from {SHA1})`
                    continue

                branch = branch.lstrip("* ")
                # Rebase the branch to the last commit from the stack .
                branches_to_rebase[branch] = [commit["node"], commit["orig-node"]]

        return branches_to_rebase

    def finalize(self, commits):
        """Rebase all branches based on changed commits from the stack."""
        branches_to_rebase = self._find_branches_to_rebase(commits)

        for branch, nodes in branches_to_rebase.items():
            self.checkout(branch)
            self._rebase(*nodes)

        self.checkout(self.branch)

    def refresh_commit_stack(self, commits):
        """Update revset and names of the commits."""
        for commit in commits:
            commit["name"] = short_node(commit["node"])
        self.revset = (commits[0]["node"], commits[-1]["node"])

    def _cherry(self, remotes):
        """Run `git cherry` and try all the remotes until success."""
        command = ["cherry", "--abbrev=12"]
        if not remotes:
            return self.git_out(command)

        for remote in remotes:
            logger.info('Determining the commit range using upstream "%s"', remote)

            try:
                response = self.git_out(command + [remote])
            except CommandError:
                continue

            return response

    def _get_first_unpublished_node(self):
        """Check which commits should be pushed and return the oldest one."""
        remotes = config.git_remote
        if self.args.upstream:
            remotes = self.args.upstream
        elif not remotes:
            remotes = self.git_out(["remote"])
            if len(remotes) > 1:
                logger.warning("!! Found multiple upstreams (%s).", ", ".join(remotes))

        unpublished = self._cherry(remotes)
        if unpublished is None:
            raise Error(
                "Unable to detect the start commit. Please provide its SHA-1 or\n"
                "specify the upstream branch with `--upstream <branch>`."
            )

        if not unpublished:
            return None

        if len(unpublished) > 100:
            raise Error(
                "Unable to create a stack with %s unpublished commits.\n\n"
                "This is usually the result of a failure to detect the correct "
                "remote repository.\nTry again with the `--upstream <upstream>` "
                "switch to specify the correct remote repository." % len(unpublished)
            )

        for line in unpublished:
            # `git cherry` is producing the output in reverse order - oldest
            # commit is the first one. That is the *opposite* of what we can find
            # in the documentation.
            if line.startswith("+"):
                return line.split("+ ")[1]
            else:
                logger.warning(
                    "!! Diff from commit %s found in upstream - omitting.",
                    line.split("- ")[1],
                )

    def set_args(self, args):
        """Store moz-phab command line args and set the revset."""
        super().set_args(args)

        self.git.set_args(args)
        if hasattr(self.args, "start_rev"):
            is_single = hasattr(self.args, "single") and self.args.single
            if self.args.start_rev == environment.DEFAULT_START_REV:
                if is_single:
                    start_rev = "HEAD"
                else:
                    start_rev = self._get_first_unpublished_node()
            else:
                start_rev = self.args.start_rev

            if start_rev is None:
                return None

            # We want inclusive range of commits if start commit is detected
            if self.args.start_rev == environment.DEFAULT_START_REV or is_single:
                start = "%s^" % start_rev
            else:
                start = start_rev

            end = start_rev if is_single else self.args.end_rev
            self.revset = (start, end)

    def _git_get_children(self, node):
        """Get commits SHA1 with their children.

        Args:
            node: The SHA1 of a node to check for all children

        Returns: A list of "aaaa bbbb cccc"" strings, where bbbb and cccc are
            SHA1 of direct children of aaaa
        """

        # Logging is disabled for this command as it can generate a _lot_ of
        # results, especially on mozilla-central.
        return self.git_out(
            ["rev-list", "--all", "--children", "--not", "%s^@" % node], never_log=True
        )

    @staticmethod
    def _get_direct_children(node, rev_list):
        """Return direct children of the commit.

        Args:
            node: The SHA1 of a node to check for direct children
            rev_list: A list of SHA1 strings - result of the _git_get_children method

        Returns: A list of SHA1 representing direct children of a commit
        """
        # Find the line containing the node to extract its commit's children
        for line in rev_list:
            if line.startswith(node):
                children = line.split(" ")
                children.remove(node)
                return children

        return []

    def _get_commits_info(self, start, end):
        """Log useful info about the commits within the desired range.

        Returns a list of strings
        An example of a list item:
            Tue, 22 Jan 2019 13:42:48 +0000
            Conduit User
            conduit@mozilla.bugs
            4912923
            b18312ffe929d3482f1d7b1e9716a1885c7a61b8
            5f161c70fef9e59d1966bab693a0a68a9336af80
            Update code references

            Fixes:
            $ moz-phab self-update
            > Failed to download update: HTTP Error 404: Not Found
        """
        boundary = "--%s--\n" % uuid.uuid4().hex
        log = self.git_out(
            [
                "log",
                "--reverse",
                "--ancestry-path",
                "--quiet",
                "--format=%aD%n%an%n%ae%n%p%n%T%n%H%n%s%n%n%b{}".format(boundary),
                "{}..{}".format(start, end),
            ],
            split=False,
            strip=False,
        )[: -len(boundary) - 1]
        return log.split("%s\n" % boundary)

    def _is_child(self, parent, node, rev_list):
        """Check if `node` is a direct or indirect child of the `parent`.

        Args:
            parent: The parent node whose children will be searched
            node: The string we check if it's in parent-child relation to the `parent`
            rev_list: A response from the git _git_get_children method - a list of
                "aaaa bbbb cccc"" strings, where "bbbb" and cccc" are SHA1 of direct
                children of "aaaa"

        Returns: a Boolean True if the `node` represents a child of the `parent`.
        """
        direct_children = self._get_direct_children(parent, rev_list)
        if node in direct_children:
            return True

        for child in direct_children:
            if self._is_child(child, node, rev_list):
                return True

        return False

    def commit_stack(self, single=False):
        """Collect all the info about commits."""
        if not self.revset:
            # No commits found to submit
            return None

        commits = []
        rev_list = None
        first_node = None
        for log_line in self._get_commits_info(*self.revset):
            if not log_line:
                continue

            (
                author_date,
                author_name,
                author_email,
                parents,
                tree_hash,
                node,
                desc,
            ) = log_line.split("\n", 6)
            desc = desc.splitlines()

            if not single:
                # Check if the commit is a child of the first one
                if rev_list is None:
                    rev_list = self._git_get_children(node)
                    first_node = node
                elif not self._is_child(first_node, node, rev_list):
                    raise Error(
                        "Commit %s is not a child of %s, unable to continue"
                        % (short_node(node), short_node(first_node))
                    )

            # Check if commit has multiple parents, if so - raise an Error
            # We may push the merging commit if it's the first one
            parents = parents.split(" ")
            if node != first_node and len(parents) > 1:
                raise Error(
                    "Multiple parents found for commit %s, unable to continue"
                    % short_node(node)
                )

            # Tue, 14 Apr 2020 12:02:20 +0000
            commit_epoch = datetime.strptime(
                author_date, "%a, %d %b %Y %H:%M:%S %z"
            ).timestamp()
            commits.append(
                {
                    "name": short_node(node),
                    "node": node,
                    "orig-node": node,
                    "submit": True,
                    "title": desc[0],
                    "title-preview": desc[0],
                    "body": "\n".join(desc[1:]).rstrip(),
                    "bug-id": None,
                    "reviewers": dict(request=[], granted=[]),
                    "rev-id": None,
                    "parent": parents[0],
                    "tree-hash": tree_hash,
                    "author-date": author_date,
                    "author-date-epoch": commit_epoch,
                    "author-name": author_name,
                    "author-email": author_email,
                }
            )

        return commits

    def is_node(self, node):
        try:
            node_type = self.git_out(
                ["cat-file", "-t", node], split=False, stderr=subprocess.STDOUT
            )
        except CommandError:
            return False

        return node_type == "commit"

    def check_node(self, node):
        """Check if the node exists.

        Calls `hg2git` if node is not found and cinnabar extension is installed.

        Returns a node if found.

        Raises NotFoundError if not found.
        """
        hashtag = node
        if not self.is_node(hashtag):
            if self.is_cinnabar_required and self.is_cinnabar_installed:
                hashtag = self._hg_to_git(hashtag)
                if hashtag == "0" * 40:
                    # hashtag is not found via hg2git
                    raise NotFoundError(
                        "Mercurial SHA1 not found by the cinnabar extension."
                    )
                elif not self.is_node(hashtag):
                    # the found hashtag is not a valid node in the repository.
                    raise NotFoundError(
                        "Mercurial SHA1 detected, but commit not found in the "
                        "repository."
                    )
            else:
                raise NotFoundError("Cinnabar extension not enabled.")

        return hashtag

    def checkout(self, node):
        self.git_call(["checkout", "--quiet", node])

    def commit(self, body, author=None, author_date=None):
        """Commit the changes in the working directory."""
        commands = ["commit", "-a"]
        if author:
            commands.append('--author="%s"' % author)

        if author_date:
            commands.append('--date="format:raw:%s 0"' % author_date)

        with temporary_file(body) as temp_f:
            commands += ["-F", temp_f]
            self.git_call(commands)

    def before_patch(self, node, name):
        """Prepare repository to apply the patches.

        Args:
            node - SHA1 of the base commit
            name - name of the branch to be created
        """
        is_detached_head = self.args.no_branch and node
        if is_detached_head and not self.args.yes:
            res = prompt(
                "Switching to the 'detached HEAD' state. Do you wish to continue?",
                ["Yes", "No"],
            )
            if res == "No":
                sys.exit(1)

        if is_detached_head and self.args.yes:
            logger.warning("Switching to the 'detached HEAD' state.")

        if is_detached_head:
            logger.warning(
                "If you want to create a new branch to retain created commits,\n"
                "you may do so by calling `git checkout -b <new-branch-name>`"
            )

        # Checkout sha
        if node:
            with wait_message("Checking out %s.." % short_node(node)):
                self.checkout(node)
            logger.info("Checked out %s", short_node(node))

        if name and not self.args.no_branch:
            branches = self.git_out(["branch", "--list", "%s*" % name])
            branches = [re.sub("[ *]", "", b) for b in branches]
            branch_name = name
            i = 0
            while branch_name in branches:
                i += 1
                branch_name = "%s_%s" % (name, i)

            self.git_call(["checkout", "-q", "-b", branch_name])
            logger.info("Created branch %s", branch_name)

    def apply_patch(self, diff, body, author, author_date):
        # apply the patch as a binary file to ensure the correct line endings
        # is used.
        with temporary_binary_file(diff.encode("utf8")) as patch_file:
            self.git_call(["apply", "--index", patch_file])

        self.commit(body, author, author_date)

    def format_patch(self, diff, body, author, author_date):
        return diff

    def _get_current_head(self):
        """Return current's HEAD symbolic link."""
        symbolic = self.git_out(["symbolic-ref", "HEAD"], split=False)
        return symbolic.split("refs/heads/")[1]

    def _get_current_hash(self):
        """Return the SHA1 of the current commit."""
        return self._revparse("HEAD")

    def _revparse(self, branch):
        """Return the SHA1 of given branch."""
        return self.git_out(["rev-parse", branch], split=False)

    def _commit_tree(
        self, parent, tree_hash, message, author_name, author_email, author_date
    ):
        """Prepare and run `commit-tree` command.

        Creates a new commit for the tree_hash.
        Args:
            parent: SHA1 of the parent commit
            tree_hash: SHA1 of the tree_hash to use for the commit
            message: commit message

        Returns:
            str: SHA1 of the new commit.
        """
        with temporary_file(message) as message_file:
            return self.git_out(
                ["commit-tree", "-p", parent, "-F", message_file, tree_hash],
                split=False,
                extra_env={
                    "GIT_AUTHOR_NAME": author_name,
                    "GIT_AUTHOR_EMAIL": author_email,
                    "GIT_AUTHOR_DATE": author_date,
                },
            )

    def amend_commit(self, commit, commits):
        """Amend the commit with an updated message.

        Changing commit's message changes also its SHA1.
        All the children within the stack and branches are then updated
        to keep the history.

        Args:
            commit: Information about the commit to be amended
            commits: List of commits within the stack
        """
        updated_body = "%s\n%s" % (commit["title"], commit["body"])

        current_body = self.git_out(
            ["show", "-s", "--format=%s%n%b", commit["node"]], split=False
        )
        if current_body == updated_body:
            logger.debug("not amending commit %s, unchanged", commit["name"])
            return

        # Create a new commit with the updated body.
        new_parent_sha = self._commit_tree(
            commit["parent"],
            commit["tree-hash"],
            updated_body,
            commit["author-name"],
            commit["author-email"],
            commit["author-date"],
        )

        # Update commit info
        commit["node"] = new_parent_sha
        # Update parent for all the children of the `commit` within the stack
        has_children = False
        for c in commits:
            if not has_children:
                # Find the amended commit info in the list of all commits in the stack.
                # Next commits are children of this one.
                has_children = c == commit
                continue

            # Update parent information and create a new commit
            c["parent"] = new_parent_sha
            new_parent_sha = self._commit_tree(
                new_parent_sha,
                c["tree-hash"],
                "%s\n%s" % (c["title"], c["body"]),
                c["author-name"],
                c["author-email"],
                c["author-date"],
            )
            c["node"] = new_parent_sha

    def rebase_commit(self, source_commit, dest_commit):
        self._rebase(dest_commit["node"], source_commit["node"])

    def is_descendant(self, node: str) -> bool:
        try:
            # See `git help merge-base` for more.
            # Note that this function is trying to determine if a commit is a
            # descendant, but `merge-base` supports checking for an ancestor. These
            # are the inverse of each other.
            self.git_out(["merge-base", "--is-ancestor", node, self.revset[0]])
        except CommandError as e:
            # Exit code 1 means the commit is not a descendant.
            if e.status == 1:
                return False

            # Any status > 1 is a command error - send that up to the user.
            if e.status > 1:
                raise e

        # If the command ran without an error, the commit is a descendant.
        return True

    def map_callsign_to_unified_head(self, callsign: str) -> Optional[str]:
        # TODO: there are apparently multiple ways to use git remote refs/branches,
        # a template may not be appropriate.
        # See https://github.com/glandium/git-cinnabar#remote-refs-styles
        unified_head = f"remotes/origin/bookmarks/{callsign}"

        if not self.is_node(unified_head):
            return None

        return unified_head

    def uplift_commits(self, dest: str, commits: List[dict]) -> List[dict]:
        # Branch name for the uplift.
        mozphab_uplift_branch = f"{self.branch}_uplift"

        # Create a new branch at the location of the tip of the revset.
        self.git_call(["switch", "-c", mozphab_uplift_branch, self.revset[-1]])

        # Rebase from the other end of the revset onto our target, specifying
        # the commit before our revset start rev as the base.
        self.git_call(["rebase", "--onto", dest, f"{self.revset[0]}^"])

        # Update revset.
        current = self._get_current_hash()
        base_rev = f"{current}~{len(commits)}"
        self.revset = self._revparse(base_rev), current

        # Get new commit stack and raise if we can't detect any - this shouldn't happen.
        new_commits = self.commit_stack()
        if not new_commits:
            raise ValueError("Didn't find any new commits after rebase!")

        # Get new commit stack and update.
        return new_commits

    def _rebase(self, newbase, upstream):
        self.git_call(["rebase", "--quiet", "--onto", newbase, upstream])

    @lru_cache(maxsize=128)
    def _file_size(self, blob):
        return int(self.git_out(["cat-file", "-s", blob], split=False))

    @lru_cache(maxsize=128)
    def _cat_file(self, blob):
        return self.git_out(["cat-file", "blob", blob], split=False, expect_binary=True)

    def _parse_diff_change(self, raw, diff):
        """Parse the changes provided in raw `git` response.

        Returns a Diff.Change object.
        """
        # find changed path
        paths = raw.split("\0")
        fields = paths.pop(0)
        [a_mode, b_mode, a_blob, b_blob, kind_l] = fields.split(" ")

        # Figure out what paths to use
        if len(paths) == 2:
            [a_path, b_path] = paths
        else:
            a_path = b_path = paths[0]

        # create a Change object
        change = diff.change_for(b_path)

        # Extract the bodies of blobs to compare
        if a_blob == NULL_SHA1:
            a_blob, a_body, a_size = None, b"", 0
        else:
            a_body = self._cat_file(a_blob)
            a_size = self._file_size(a_blob)

        if b_blob == NULL_SHA1:
            b_blob, b_body, b_size = None, b"", 0
        else:
            b_body = self._cat_file(b_blob)
            b_size = self._file_size(b_blob)

        file_size = max(a_size, b_size)
        telemetry().submission.files_size.accumulate(file_size)

        # Detect if we're binary, and generate a unified diff
        if b"\0" in a_body or b"\0" in b_body or file_size > environment.MAX_TEXT_SIZE:
            change.binary = True

        if not change.binary and a_body:
            try:
                a_body = str(a_body, "utf-8")
            except UnicodeDecodeError:
                change.binary = True

        if not change.binary and b_body:
            try:
                b_body = str(b_body, "utf-8")
            except UnicodeDecodeError:
                change.binary = True

        if change.binary:
            change.set_as_binary(
                a_body=a_body,
                a_mime=mimetypes.guess_type(a_path)[0] or "",
                b_body=b_body,
                b_mime=mimetypes.guess_type(b_path)[0] or "",
            )

        else:
            # We can only diff changed blobs.
            if a_blob == b_blob:
                # No changes in the file contents.
                lines = create_hunk_lines(a_body, " ", False)[0]
                if lines:
                    change.hunks.append(
                        Diff.Hunk(
                            old_off=1,
                            old_len=len(lines),
                            new_off=1,
                            new_len=len(lines),
                            lines=lines,
                        )
                    )
            elif a_blob is None:
                # The file is created.
                lines, eof_missing_newline = create_hunk_lines(b_body, "+")
                if lines:
                    new_len = len(lines)
                    if eof_missing_newline:
                        new_len -= 1
                    change.hunks.append(
                        Diff.Hunk(
                            old_off=0,
                            old_len=0,
                            new_off=1,
                            new_len=new_len,
                            lines=lines,
                        )
                    )

            elif b_blob is None and file_size:
                # The file is removed.
                lines, eof_missing_newline = create_hunk_lines(a_body, "-")
                if lines:
                    old_len = len(lines)
                    if eof_missing_newline:
                        old_len -= 1
                    change.hunks.append(
                        Diff.Hunk(
                            old_off=1,
                            old_len=old_len,
                            new_off=0,
                            new_len=0,
                            lines=lines,
                        )
                    )
            elif b_blob is not None:
                # There are changes in the file.
                if self.args.lesscontext or file_size > environment.MAX_CONTEXT_SIZE:
                    context_size = 100
                else:
                    context_size = environment.MAX_CONTEXT_SIZE

                diff_args = [
                    "diff",
                    "--submodule=short",
                    "--no-ext-diff",
                    "--no-color",
                    "--no-textconv",
                    "-U%s" % context_size,
                    a_blob,
                    b_blob,
                ]
                git_diff = self.git_out(diff_args, expect_binary=True).decode("utf-8")
                change.from_git_diff(git_diff)

        diff.set_change_kind(change, kind_l[0], a_mode, b_mode, a_path, b_path)

        return change

    def get_diff(self, commit):
        """Create a Diff object with changes."""
        raw = self.git_out(
            [
                "diff-tree",
                "-r",
                "--raw",
                "-z",
                "-M",
                "-C",
                "--no-abbrev",
                commit["node"],
            ],
            split=False,
        )

        diff = Diff()
        for raw_change in raw[:-1].split("\0:")[1:]:
            self._parse_diff_change(raw_change, diff)

        return diff

    def check_vcs(self):
        if self.args.force_vcs:
            return True

        if self.is_cinnabar_required and not self.is_cinnabar_installed:
            raise Error(
                "Git Cinnabar extension is required to work on this repository."
            )

        return True

    def validate_email(self):
        """Validate the user's configured email (user.email)."""
        if not is_valid_email(self.git.email):
            raise Error(
                f"Your email configured with git ({self.git.email}) is not a valid "
                f"format.\nPlease run `git config user.email some@email.com` to set "
                f"the correct value.\n\nYou can also run `git commit --amend "
                f'--author="Author Name <some@email.com>" --no-edit` to amend '
                f"the most recent commit."
            )



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pytest
from pathlib import Path
from unittest import mock

from .conftest import create_temp_fn

from mozphab import environment, exceptions, mozphab


@mock.patch("mozphab.git.Git.git_out")
def test_cherry(m_git_git_out, git):
    m_git_git_out.side_effect = (exceptions.CommandError, ["output"])
    assert git._cherry(["one", "two"]) == ["output"]
    m_git_git_out.assert_has_calls(
        [
            mock.call(["cherry", "--abbrev=12", "one"]),
            mock.call(["cherry", "--abbrev=12", "two"]),
        ]
    )


@mock.patch("mozphab.git.Git.git_out")
@mock.patch("mozphab.git.Git._cherry")
@mock.patch("mozphab.git.config")
def test_first_unpublished(m_config, m_git_cherry, m_git_git_out, git):
    class Args:
        def __init__(self, upstream=None, start_rev="(auto)"):
            self.upstream = upstream
            self.start_rev = start_rev

    m_config.git_remote = []
    m_git_git_out.side_effect = (["a", "b"], ["c"], ["d"])
    m_git_cherry.side_effect = (["- sha1", "+ sha2"], [], None)
    git.args = Args()
    first = git._get_first_unpublished_node
    assert "sha2" == first()
    m_git_cherry.assert_called_with(["a", "b"])
    assert first() is None
    with pytest.raises(exceptions.Error):
        first()

    m_git_cherry.side_effect = ([],)
    git.args = Args(upstream=["upstream"])
    first()
    m_git_cherry.assert_called_with(["upstream"])

    m_git_cherry.side_effect = ([],)
    m_config.git_remote = ["someremote"]
    git.args = Args()
    first()
    m_git_cherry.assert_called_with(["someremote"])
    m_config.git_remote = []

    m_git_cherry.side_effect = (["+ %s" % i for i in range(101)],)
    m_git_git_out.side_effect = (["origin"],)
    with pytest.raises(exceptions.Error):
        first()


@mock.patch("mozphab.git.Git.git_out")
def test_branches_to_rebase(m_git_git_out, git):
    git_find = git._find_branches_to_rebase

    # No branch returned - not a real case - we don't work without branches
    m_git_git_out.return_value = []
    assert dict() == git_find([{"orig-node": "_aaa", "node": "aaa"}])

    # No amend, no branches to rebase
    m_git_git_out.return_value = ["branch"]
    assert dict() == git_find([{"orig-node": "aaa", "node": "aaa"}])

    # One commit, one branch
    m_git_git_out.return_value = ["branch"]
    assert dict(branch=["aaa", "_aaa"]) == git_find(
        [{"orig-node": "_aaa", "node": "aaa"}]
    )

    # Two commits one branch
    m_git_git_out.return_value = ["branch"]
    assert dict(branch=["bbb", "_bbb"]) == git_find(
        [{"orig-node": "_aaa", "node": "aaa"}, {"orig-node": "_bbb", "node": "bbb"}]
    )

    # Two branches one commit
    # ... (branch1)
    # | ... (branch2)
    # |/
    # * aaa
    # More realistic output from the git command
    m_git_git_out.return_value = ["*  branch1", "  branch2"]
    assert dict(branch1=["aaa", "_aaa"], branch2=["aaa", "_aaa"]) == git_find(
        [{"orig-node": "_aaa", "node": "aaa"}]
    )

    # ... (branch1)
    # | * bbb (branch2)
    # |/
    # * aaa
    m_git_git_out.side_effect = (["branch1", "branch2"], ["branch2"])
    assert dict(branch1=["aaa", "_aaa"], branch2=["bbb", "_bbb"]) == git_find(
        [{"orig-node": "_aaa", "node": "aaa"}, {"orig-node": "_bbb", "node": "bbb"}]
    )

    # * ... (master)
    # | * ... (feature1)
    # | | * ... (feature2)
    # | |/
    # |/|
    # | | * ddd (feature1_1)
    # | |/
    # | * ccc
    # |/
    # * bbb
    # * aaa

    m_git_git_out.side_effect = (
        ["master", "feature1", "feature1_1", "feature2"],  # aaa
        ["master", "feature1", "feature1_1", "feature2"],  # bbb
        ["feature1", "feature1_1"],  # ccc
        ["feature1_1"],  # ddd
    )
    assert dict(
        master=["bbb", "_bbb"],
        feature1=["ccc", "_ccc"],
        feature2=["bbb", "_bbb"],
        feature1_1=["ddd", "_ddd"],
    ) == git_find(
        [
            {"orig-node": "_aaa", "node": "aaa"},
            {"orig-node": "_bbb", "node": "bbb"},
            {"orig-node": "_ccc", "node": "ccc"},
            {"orig-node": "_ddd", "node": "ddd"},
        ]
    )


def test_get_direct_children(git):
    get_children = git._get_direct_children
    rev_list = ["aaa bbb ccc", "bbb", "ccc ddd"]
    assert ["bbb", "ccc"] == get_children("aaa", rev_list)
    assert [] == get_children("bbb", rev_list)
    assert ["ddd"] == get_children("ccc", rev_list)
    assert [] == get_children("xxx", rev_list)


def test_is_child(git):
    is_child = git._is_child
    # * ccc
    # * bbb
    # * aaa
    nodes = ["ccc", "bbb ccc", "aaa bbb"]
    assert is_child("aaa", "bbb", nodes)
    assert is_child("aaa", "ccc", nodes)
    assert is_child("bbb", "ccc", nodes)
    assert not is_child("bbb", "aaa", nodes)
    assert not is_child("aaa", "aaa", nodes)
    assert not is_child("bbb", "bbb", nodes)
    assert not is_child("ccc", "ccc", nodes)

    # * ddd
    # | * ccc
    # | | * eee
    # | |/
    # | * bbb
    # |/
    # * aaa
    nodes = ["ddd", "ccc", "eee", "bbb ccc eee", "aaa bbb ddd"]
    assert is_child("aaa", "bbb", nodes)
    assert is_child("aaa", "ccc", nodes)
    assert is_child("aaa", "ddd", nodes)
    assert is_child("aaa", "eee", nodes)
    assert is_child("bbb", "ccc", nodes)
    assert is_child("bbb", "eee", nodes)
    assert not is_child("bbb", "ddd", nodes)
    assert not is_child("ccc", "ddd", nodes)


@mock.patch("mozphab.git.Git.git_out")
@mock.patch("mozphab.mozphab.config")
def test_range(m_config, m_git_git_out, git):
    class Args:
        def __init__(self, start="start", end="."):
            self.start_rev = start
            self.end_rev = end
            self.safe_mode = False
            self.single = False

    m_config.safe_mode = False
    m_git_git_out.return_value = ["user.email=email"]
    git.set_args(Args())
    assert git.revset == ("start", ".")


@mock.patch("mozphab.mozphab.config")
@mock.patch("mozphab.gitcommand.parse_config")
@mock.patch("mozphab.git.Git._get_first_unpublished_node")
@mock.patch("mozphab.git.Git.git_out")
def test_set_args(m_git_git_out, m_git_get_first, m_parse_config, m_config, git):
    class Args:
        def __init__(
            self,
            start=environment.DEFAULT_START_REV,
            end=environment.DEFAULT_END_REV,
            safe_mode=False,
            single=False,
        ):
            self.start_rev = start
            self.end_rev = end
            self.safe_mode = safe_mode
            self.single = single

    with pytest.raises(exceptions.Error):
        git.set_args(Args())

    git.git.command = ["git"]
    git.git.safe_mode = False
    m_parse_config.return_value = {"user.email": "email"}
    m_git_get_first.return_value = "aaa"
    git.set_args(Args())
    assert ["git"] == git.git.command
    m_git_get_first.assert_called_once()
    assert git.revset == ("aaa^", ".")

    m_parse_config.return_value = {
        "user.email": "email",
        "user.name": "name",
        "cinnabar.helper": "string",
    }
    git.set_args(Args())
    assert ["git"] == git.git.command
    assert ["cinnabar"] == git.git.extensions

    safe_options = (
        ["git"]
        + ["-c", "user.email=email"]
        + ["-c", "user.name=name"]
        + ["-c", "cinnabar.helper=string"]
    )
    git.set_args(Args(safe_mode=True))
    assert safe_options == git.git.command

    git.git.command = ["git"]
    git.git.safe_mode = True
    git.set_args(Args())
    assert safe_options == git.git.command

    m_config.safe_mode = False
    m_git_get_first.reset_mock()
    git.set_args(Args(start="bbb", end="ccc"))
    m_git_get_first.assert_not_called()
    assert git.revset == ("bbb", "ccc")

    git.set_args(Args(safe_mode=True))
    assert "" == git.git._env["HOME"]
    assert "" == git.git._env["XDG_CONFIG_HOME"]

    m_config.safe_mode = True
    git.set_args(Args())
    assert "" == git.git._env["HOME"]
    assert "" == git.git._env["XDG_CONFIG_HOME"]

    m_git_get_first.reset_mock()
    m_parse_config.return_value = {
        "user.email": "email",
        "user.name": "name",
    }
    git.set_args(Args(single=True))
    m_git_get_first.assert_not_called()
    assert git.revset == ("HEAD^", "HEAD")

    git.set_args(Args(single=True, start="start"))
    m_git_get_first.assert_not_called()
    assert git.revset == ("start^", "start")


@mock.patch("mozphab.git.Git.git_out")
def test_worktree_clean(m_git_out, git):
    m_git_out.return_value = ""
    assert git.is_worktree_clean()

    m_git_out.return_value = ["xxx"]
    assert not git.is_worktree_clean()

    m_git_out.return_value = ["?? one", "?? two"]
    assert git.is_worktree_clean()

    m_git_out.return_value = ["?? one", "?? two", " M xxx"]
    assert not git.is_worktree_clean()


@mock.patch("mozphab.git.Git.git_call")
def test_commit(m_git, git):
    git.commit("some body")
    assert m_git.called_once()

    m_git.reset_mock()
    git.commit("some body", "user")
    assert m_git.called_once()


@mock.patch("mozphab.git.Git._hg_to_git")
@mock.patch("mozphab.git.Git.is_node")
@mock.patch("mozphab.git.Git.phab_vcs")
def test_check_node(m_phab_vcs, m_git_is_node, m_hg2git, git):
    node = "aabbcc"
    mozphab.conduit.set_repo(git)
    assert node == git.check_node(node)

    git._phab_vcs = "hg"
    git.git._cinnabar_installed = False
    m_git_is_node.return_value = False
    with pytest.raises(exceptions.NotFoundError) as e:
        git.check_node(node)
    assert "Cinnabar extension not enabled" in str(e.value)

    git.git._cinnabar_installed = True
    m_hg2git.return_value = "0" * 40
    with pytest.raises(exceptions.NotFoundError) as e:
        git.check_node(node)
    assert "Mercurial SHA1 not found" in str(e.value)

    m_hg2git.return_value = "git_aabbcc"
    with pytest.raises(exceptions.NotFoundError) as e:
        git.check_node(node)
    assert "Mercurial SHA1 detected, but commit not found" in str(e.value)

    m_git_is_node.side_effect = (False, True)
    assert "git_aabbcc" == git.check_node(node)


@mock.patch("mozphab.git.Git.git_out")
@mock.patch("mozphab.git.Git.checkout")
@mock.patch("mozphab.git.Git.git_call")
@mock.patch("mozphab.git.prompt")
@mock.patch("mozphab.git.logger")
def test_before_patch(m_logger, m_prompt, m_git, m_checkout, m_git_out, git):
    class Args:
        def __init__(
            self,
            rev_id="D123",
            nocommit=False,
            raw=False,
            applyto="base",
            no_branch=False,
            yes=False,
        ):
            self.rev_id = rev_id
            self.nocommit = nocommit
            self.raw = raw
            self.applyto = applyto
            self.no_branch = no_branch
            self.yes = yes

    git.args = Args()
    m_git_out.side_effect = (["  branch"],)
    git.before_patch("sha1", "branch")
    m_checkout.assert_called_with("sha1")
    m_git.assert_called_with(["checkout", "-q", "-b", "branch_1"])

    m_git.reset_mock()
    m_git_out.reset_mock()
    m_checkout.reset_mock()

    m_checkout.reset_mock()
    m_git_out.side_effect = ("the branch name is not here",)
    git.args = Args(applyto="here")
    git.before_patch(None, "branchname")
    m_checkout.assert_not_called()

    m_git.reset_mock()
    m_checkout.reset_mock()
    git.args = Args(applyto="abcdef", nocommit=True)
    git.before_patch("abcdef", None)
    m_checkout.assert_called_once_with("abcdef")
    m_git.assert_not_called()

    m_git.reset_mock()
    m_checkout.reset_mock()
    m_logger.reset_mock()
    git.args = Args(no_branch=True, yes=True)
    git.before_patch("abcdef", "name")
    m_checkout.assert_called_once()
    m_git.assert_not_called()
    assert "git checkout -b" in m_logger.warning.call_args_list[1][0][0]

    m_git.reset_mock()
    m_checkout.reset_mock()
    m_logger.reset_mock()
    git.args = Args(no_branch=True)
    git.before_patch("abcdef", "name")
    m_checkout.assert_called_once()
    m_git.assert_not_called()
    m_prompt.assert_called_once()
    assert "git checkout -b" in m_logger.warning.call_args_list[0][0][0]

    m_prompt.return_value = "No"
    with pytest.raises(SystemExit):
        git.before_patch("abcdef", "name")


@mock.patch("mozphab.git.temporary_binary_file")
@mock.patch("mozphab.git.Git.git_call")
@mock.patch("mozphab.git.Git.commit")
def test_apply_patch(m_commit, m_git, m_temp_fn, git):
    m_temp_fn.return_value = create_temp_fn("filename")
    git.apply_patch("diff", "commit message", "user", 1)
    m_git.assert_called_once_with(["apply", "--index", "filename"])
    m_commit.assert_called_with("commit message", "user", 1)
    m_temp_fn.assert_called_once_with(b"diff")


@mock.patch("mozphab.git.Git.git_out")
def test_is_node(m_git_out, git):
    m_git_out.return_value = "commit"
    assert git.is_node("aaa")

    m_git_out.return_value = "something else"
    assert not git.is_node("aaa")

    m_git_out.side_effect = exceptions.CommandError
    assert not git.is_node("aaa")


@mock.patch("mozphab.git.Git.git_out")
def test_is_descendant(m_git_out, git):
    git.revset = ("abc", "def")

    m_git_out.return_value = ""
    assert (
        git.is_descendant("aabbcc") is True
    ), "Call to merge-base with 0 return code should indicate node is a descendant."

    m_git_out.side_effect = exceptions.CommandError("", 1)
    assert (
        git.is_descendant("aabbcc") is False
    ), "Call to merge-base with 1 return code should indicated node is not descendant."

    m_git_out.side_effect = exceptions.CommandError("test", 255)
    with pytest.raises(exceptions.CommandError) as e:
        git.is_descendant("aabbcc")

        assert (
            e.args == "test"
        ), "Original command exception was not raised to the caller."


@mock.patch("mozphab.gitcommand.which")
@mock.patch("mozphab.gitcommand.GitCommand.output")
def test_is_cinnabar_installed(m_git_out, m_which, git, tmp_path):
    def _without_str(calls):
        # Debuggers call __str__ on mocked functions, strip them
        return [c for c in calls if c[0] != "__str__"]

    # cinnabar installed as visible external command
    m_git_out.return_value = ["External commands", "   cinnabar", "   revise"]
    assert git.is_cinnabar_installed
    m_git_out.assert_called_once_with(["help", "--all"])

    # cached request (primed by 'cinnabar installed' test)
    m_git_out.reset_mock()
    m_git_out.return_value = ["External commands", "   cinnabar", "   revise"]
    assert git.is_cinnabar_installed
    m_git_out.assert_not_called()

    # create a fake cinnabar in exec-path for git to find
    cinnabar = Path(tmp_path) / "git-cinnabar"
    cinnabar.write_text("")
    cinnabar.chmod(0o755)

    # cinnabar installed in exec-path
    m_git_out.reset_mock()
    m_git_out.side_effect = [
        ["External commands", "   revise"],  # git help --all
        tmp_path,  # git --exec-path
    ]
    git.git._cinnabar_installed = None
    assert git.is_cinnabar_installed
    assert _without_str(m_git_out.mock_calls) == [
        mock.call(["help", "--all"]),
        mock.call(["--exec-path"], split=False),
    ]

    # remove cinnabar from exec-path so we fall back to looking on the path
    cinnabar.unlink()

    # cinnabar installed somewhere on path
    m_git_out.reset_mock()
    m_git_out.side_effect = [
        ["External commands", "   revise"],  # git help --all
        tmp_path,  # git --exec-path
    ]
    m_which.return_value = str(cinnabar)
    git.git._cinnabar_installed = None
    assert git.is_cinnabar_installed
    assert _without_str(m_git_out.mock_calls) == [
        mock.call(["help", "--all"]),
        mock.call(["--exec-path"], split=False),
    ]

    # cinnabar not installed
    m_git_out.reset_mock()
    m_git_out.side_effect = [
        ["External commands", "   revise"],  # git help --all
        tmp_path,  # git --exec-path
    ]
    m_which.return_value = None
    git.git._cinnabar_installed = None
    assert not git.is_cinnabar_installed
    assert _without_str(m_git_out.mock_calls) == [
        mock.call(["help", "--all"]),
        mock.call(["--exec-path"], split=False),
    ]


@mock.patch("mozphab.git.Git.git_out")
def test_unicode_in_windows_env(m_git_out, git, monkeypatch):
    monkeypatch.setattr(environment, "IS_WINDOWS", True)
    git._commit_tree("parent", "tree_hash", "message", "ćwikła", "ćwikła", "date")
    m_git_out.assert_called_once_with(
        ["commit-tree", "-p", "parent", "-F", mock.ANY, "tree_hash"],
        split=False,
        extra_env=dict(
            GIT_AUTHOR_NAME="ćwikła", GIT_AUTHOR_EMAIL="ćwikła", GIT_AUTHOR_DATE="date"
        ),
    )


def test_check_vcs(git):
    class Args:
        def __init__(self, force_vcs=False):
            self.safe_mode = False
            self.force_vcs = force_vcs

    args = Args()
    git.set_args(args)
    assert git.check_vcs()

    git.git._cinnabar_installed = True
    git._phab_vcs = "hg"
    assert git.check_vcs()

    git.git._cinnabar_installed = False
    with pytest.raises(exceptions.Error):
        git.check_vcs()

    args = Args(force_vcs=True)
    git.set_args(args)
    assert git.check_vcs()


@mock.patch("mozphab.git.Git._get_commits_info")
@mock.patch("mozphab.git.Git._git_get_children")
@mock.patch("mozphab.git.Git._is_child")
def test_commit_stack_single(_1, _2, _3, git):
    git._get_commits_info.return_value = [
        """\
Tue, 22 Jan 2019 13:42:48 +0000
Conduit User
conduit@mozilla.bugs
4912923
b18312ffe929d3482f1d7b1e9716a1885c7a61b8
aaa000aaa000aaa000aaa000aaa000aaa000aaa0
title

description
"""
    ]
    git.revset = ["HEAD^", "HEAD"]
    git.commit_stack(single=True)
    git._git_get_children.assert_not_called()
    git._is_child.assert_not_called()


@mock.patch("mozphab.git.Git.is_node")
def test_git_map_callsign_to_unified_head(m_is_node, git):
    # If head is not a node in the repo, raise `ValueError`.
    m_is_node.return_value = False
    assert (
        git.map_callsign_to_unified_head("beta") is None
    ), "Unknown head should have returned `None`."

    # If head is a node in the repo, should map to a remote branch.
    m_is_node.return_value = True
    assert (
        git.map_callsign_to_unified_head("beta") == "remotes/origin/bookmarks/beta"
    ), "beta did not correctly map to a branch"


def test_git_validate_email(git):
    with pytest.raises(exceptions.Error):
        git.validate_email()

    git.git.email = "test@mozilla.com"
    assert git.validate_email() is None, "validate_email() executes without error"



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import calendar
import contextlib
import logging
import logging.handlers
import os
import re
import sys
import time

from glob import glob

from mozphab import environment

from colorama import just_fix_windows_console

logger = logging.getLogger("moz-phab")


LOG_MAX_SIZE = 1024 * 1024 * 50
LOG_BACKUPS = 5


_handlers = list()


class ColourFormatter(logging.Formatter):
    def __init__(self):
        if environment.DEBUG:
            fmt = "%(levelname)-8s %(asctime)-13s %(message)s"
        else:
            fmt = "%(message)s"
        super().__init__(fmt)
        self.log_colours = {"WARNING": 34, "ERROR": 31}  # blue, red
        # wrap stdout and stderr, so that color codes work in the windows console
        just_fix_windows_console()

    def format(self, record):
        result = super().format(record)
        if environment.HAS_ANSI and record.levelname in self.log_colours:
            result = "\033[%sm%s\033[0m" % (self.log_colours[record.levelname], result)
        return result


def init_logging():
    """Initialize logging."""
    log_file = os.path.join(environment.MOZBUILD_PATH, "moz-phab.log")
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(ColourFormatter())
    handler.setLevel(logging.DEBUG if environment.DEBUG else logging.INFO)
    logger.addHandler(handler)
    _handlers.append(handler)

    handler = logging.handlers.RotatingFileHandler(
        filename=log_file,
        maxBytes=LOG_MAX_SIZE,
        backupCount=LOG_BACKUPS,
        encoding="utf8",
    )
    handler.setFormatter(
        logging.Formatter("%(asctime)-13s %(levelname)-8s %(message)s")
    )
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    _handlers.append(handler)

    logger.setLevel(logging.DEBUG)

    # clean up old date-based logs
    now = time.time()
    for filename in sorted(glob("%s/*.log.*" % os.path.dirname(log_file))):
        m = re.search(r"\.(\d\d\d\d)-(\d\d)-(\d\d)$", filename)
        if not m:
            continue
        file_time = calendar.timegm(
            (int(m.group(1)), int(m.group(2)), int(m.group(3)), 0, 0, 0)
        )
        if (now - file_time) / (60 * 60 * 24) > 8:
            logger.debug("deleting old log file: %s" % os.path.basename(filename))
            with contextlib.suppress(IOError):
                os.unlink(filename)


def stop_logging():
    """Remove our logging handlers and close files."""

    for handler in _handlers:
        logger.removeHandler(handler)
        if isinstance(handler, logging.FileHandler):
            handler.close()



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mozphab import updater


def test_parse_latest_prerelease_version():
    # Test data from the `simple` api.
    data = {
        "files": [
            {
                "filename": "MozPhab-1.2.2rc0.tar.gz",
            },
            {
                "filename": "MozPhab-1.2.2rc1-py3-none-any.whl",
            },
            {
                "filename": "MozPhab-1.2.2rc1.tar.gz",
            },
            {
                "filename": "MozPhab-1.2.0.tar.gz",
            },
        ],
    }

    assert (
        updater.parse_latest_prerelease_version(data) == "1.2.2rc1"
    ), "`get_newest_pypi_version` should detect `1.2.2rc1` as the latest version."
